"""
Copyright 2017 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os

from bzt import TaurusConfigError, TaurusInternalException
from bzt.modules import SubprocessedExecutor
from bzt.utils import get_full_path, is_windows, RequiredTool, RESOURCES_DIR, CALL_PROBLEMS


class CSharpExecutor(SubprocessedExecutor):
    def __init__(self):
        super(CSharpExecutor, self).__init__()
        self.runner_dir = os.path.join(RESOURCES_DIR, "dotnet", "DotnetTestRunner")
        self.dotnet = None
        self.executor_name = None
        self.runner_executable = "dotnet"

    def install_required_tools(self):
        if is_windows():
            return

        self.dotnet = self._get_tool(Dotnet)
        self.log.debug("Checking for dotnet")
        if not self.dotnet.check_if_installed():
            self.dotnet.install()

    def prepare(self):
        super(CSharpExecutor, self).prepare()
        self.env.set({"CopyRetryCount": 100})
        self.script = get_full_path(self.get_script_path())
        if not self.script:
            raise TaurusConfigError("Script not passed to runner %s" % self)

        self.install_required_tools()
        self.reporting_setup(suffix=".ldjson")

    def startup(self):
        if not self.executor_name:
            raise TaurusInternalException("C# executor is not specified, use NUnit or XUnit instead.")

        cmdline = []

        cmdline += [self.runner_executable, "run", "--project", self.runner_dir, "--", self.executor_name,
                    "--target", self.script,
                    "--report-file", self.report_file]

        load = self.get_load()
        if load.iterations:
            cmdline += ['--iterations', str(load.iterations)]
        if load.hold:
            cmdline += ['--duration', str(int(load.hold))]
        if load.concurrency:
            cmdline += ['--concurrency', str(int(load.concurrency))]
        if load.ramp_up:
            cmdline += ['--ramp-up', str(int(load.ramp_up))]

        self.process = self._execute(cmdline)


class NUnitExecutor(CSharpExecutor):
    def __init__(self):
        super(NUnitExecutor, self).__init__()
        self.executor_name = "nUnit"


class XUnitExecutor(CSharpExecutor):
    def __init__(self):
        super(XUnitExecutor, self).__init__()
        self.executor_name = "xUnit"


class Dotnet(RequiredTool):
    def __init__(self, **kwargs):
        super(Dotnet, self).__init__(tool_path="dotnet", installable=False, **kwargs)

    def check_if_installed(self):
        self.log.debug('Trying %s: %s', self.tool_name, self.tool_path)
        try:
            out, err = self.call([self.tool_path, '--version'])
        except CALL_PROBLEMS as exc:
            self.log.warning("%s check failed: %s", self.tool_name, exc)
            return False

        self.log.debug("%s check stdout: %s", self.tool_name, out)
        if err:
            self.log.warning("%s check stderr: %s", self.tool_name, err)
        return True



"""
Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from subprocess import CalledProcessError, PIPE

from bzt import TaurusConfigError, RCProvider
from bzt.engine import Service
from bzt.utils import ensure_is_dict, Environment, shutdown_process, shell_exec, communicate


class ShellExecutor(Service):
    """
    :type env: Environment
    :type prepare_tasks: list[Task]
    :type startup_tasks: list[Task]
    :type check_tasks: list[Task]
    :type shutdown_tasks: list[Task]
    :type postprocess_tasks: list[Task]
    """

    def __init__(self):
        super(ShellExecutor, self).__init__()
        self.prepare_tasks = []
        self.startup_tasks = []
        self.check_tasks = []
        self.shutdown_tasks = []
        self.postprocess_tasks = []
        self.env = Environment(self.log)

    def _load_tasks(self, stage, container):
        if not isinstance(self.parameters.get(stage, []), list):
            self.parameters[stage] = [self.parameters[stage]]

        for index, stage_task in enumerate(self.parameters.get(stage, [])):
            stage_task = ensure_is_dict(self.parameters[stage], index, "command")
            task_config = self.parameters[stage][index]
            default_cwd = self.settings.get("default-cwd", None)
            cwd = self.engine.find_file(task_config.get("cwd", default_cwd))
            if cwd is None:
                working_dir = self.engine.default_cwd
            elif cwd == 'artifacts-dir':
                working_dir = self.engine.artifacts_dir
            else:
                working_dir = cwd

            # make copy of env for every task
            env = Environment(self.log, self.env)
            env.set(task_config.get('env'))
            env.add_path({"PYTHONPATH": working_dir})

            task = Task(task_config, self.log, working_dir, env)

            f_name = 'shellexec_%s_%s' % (stage, index)
            if task.out:
                task.out = open(task.out, 'at')
            else:
                if task.is_background:
                    out = self.engine.create_artifact(f_name, '.out')
                    self.log.debug('STDOUT of background task "%s" redirected to "%s"' % (task, out))
                    task.out = open(out, 'at')
                else:
                    task.out = PIPE

            if task.err:
                task.err = open(task.err, 'at')
            else:
                if task.is_background:
                    err = self.engine.create_artifact(f_name, '.err')
                    self.log.debug('STDERR of background task "%s" redirected to "%s"' % (task, err))
                    task.err = open(err, 'at')
                else:
                    task.err = PIPE

            container.append(task)
            self.log.debug("Added %s task: %s", stage, stage_task)

    def prepare(self):
        """
        Configure Tasks
        :return:
        """
        self.env.set(self.settings.get('env'))

        self._load_tasks('prepare', self.prepare_tasks)
        self._load_tasks('startup', self.startup_tasks)
        self._load_tasks('check', self.check_tasks)
        self._load_tasks('shutdown', self.shutdown_tasks)
        self._load_tasks('post-process', self.postprocess_tasks)

        for task in self.prepare_tasks:
            task.start()

    def startup(self):
        for task in self.startup_tasks:
            task.start()

    def check(self):
        for task in self.check_tasks:
            task.start()

        for task in self.prepare_tasks + self.startup_tasks + self.check_tasks:
            task.check()

        return super(ShellExecutor, self).check()

    def shutdown(self):
        for task in self.shutdown_tasks:
            self._set_stop_reason_vars(task)
            task.start()

        for task in self.check_tasks + self.startup_tasks:
            task.shutdown()

    def post_process(self):
        for task in self.shutdown_tasks + self.check_tasks + self.startup_tasks + self.prepare_tasks:
            task.shutdown()

        for task in self.postprocess_tasks:
            self._set_stop_reason_vars(task)
            task.start()
            task.shutdown()

    def _set_stop_reason_vars(self, task):
        if isinstance(self.engine.stopping_reason, RCProvider):
            rc = self.engine.stopping_reason.get_rc()
            task.env.set({
                "TAURUS_STOPPING_REASON": str(self.engine.stopping_reason)
            })
        elif self.engine.stopping_reason:
            rc = 1
        else:
            rc = 0

        task.env.set({
            "TAURUS_EXIT_CODE": rc,
            "TAURUS_STOPPING_REASON": str(self.engine.stopping_reason)
        })


class Task(object):
    """
    :type process: subprocess.Popen
    :type env: Environment
    """

    def __init__(self, config, parent_log, working_dir, env):
        """
        :type env: Environment
        """
        self.log = parent_log.getChild(self.__class__.__name__)
        self.working_dir = working_dir
        self.env = env

        self.command = config.get("command", TaurusConfigError("Parameter is required: command"))
        self.is_background = config.get("background", False)

        self.out = config.get("out", None)
        self.err = config.get("err", None)

        self.ignore_failure = config.get("ignore-failure", False)
        self.process = None

    def start(self):
        """
        Start task
        """
        if self.process:
            self.check()
            self.log.info("Process already running: %s", self)
            return

        self.log.info("Starting shell command: %s", self)
        self.process = shell_exec(args=self.command, stdout=self.out, stderr=self.err, cwd=self.working_dir,
                                  env=self.env.get(), shell=True)
        if self.is_background:
            self.log.debug("Task started, PID: %d", self.process.pid)
        else:
            self._get_results()

    def check(self):
        if not self.process or self.process.returncode is not None:  # not started or already finished task
            return True

        if self.process.poll() is None:
            self.log.debug('Task is not finished yet: %s', self)
            return False

        self._get_results()
        return True

    def _get_results(self):
        stdout, stderr = communicate(self.process)

        if stdout and (self.out == PIPE):
            self.log.debug("Output for %s:\n%s", self, stdout)

        if stderr and (self.err == PIPE):
            self.log.warning("Errors for %s:\n%s", self, stderr)

        self.log.debug("Task was finished with exit code %s: %s", self.process.returncode, self)
        if not self.ignore_failure and self.process.returncode != 0:
            if self.out != PIPE:
                self.log.warning("Output for %s:\n%s", self, stdout)
            raise CalledProcessError(self.process.returncode, self)

    def shutdown(self):
        """
        If task was not completed, kill process, provide output
        else provide output
        :return:
        """
        self.check()

        if self.process and self.process.returncode is None:
            self.log.info("Background task was not completed, shutting it down: %s", self)
            shutdown_process(self.process, self.log)

        self.process = None
        for stream in (self.out, self.err):
            if stream and stream != PIPE:
                stream.close()

    def __repr__(self):
        return self.command



"""
Module for reporting into http://www.blazemeter.com/ service

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
from collections import namedtuple
from ssl import SSLError
from urllib.error import URLError

from requests.exceptions import ReadTimeout

from bzt import TaurusNetworkError

NETWORK_PROBLEMS = (IOError, URLError, SSLError, ReadTimeout, TaurusNetworkError)


def parse_blazemeter_test_link(link):
    """
    https://a.blazemeter.com/app/#/accounts/97961/workspaces/89846/projects/229969/tests/5823512

    :param link:
    :return:
    """
    if not isinstance(link, str):
        return None

    regex = r'https://a.blazemeter.com/app/#/accounts/(\d+)/workspaces/(\d+)/projects/(\d+)/tests/(\d+)(?:/\w+)?'
    match = re.match(regex, link)
    if match is None:
        return None

    TestParams = namedtuple('TestParams', 'account_id,workspace_id,project_id,test_id')
    return TestParams(*[int(x) for x in match.groups()])



"""
Module for reporting into http://www.blazemeter.com/ service

Copyright 2015 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import copy
import logging
import os
import sys
import time
import traceback
import zipfile
from collections import Counter
from time import sleep
from urllib.error import URLError

import requests
from terminaltables3 import SingleTable, AsciiTable
from urwid import Pile, Text

from bzt import AutomatedShutdown, TaurusConfigError, TaurusException, TaurusNetworkError, NormalShutdown
from bzt.bza import User, Session, BZA_TEST_DATA_RECEIVED, ENDED
from bzt.engine import Reporter, Provisioning, Configuration, Service, SETTINGS, ScenarioExecutor, EXEC
from bzt.modules.aggregator import DataPoint, KPISet, ResultsProvider
from bzt.modules.console import PrioritizedWidget
from bzt.modules.functional import FunctionalResultsReader, FunctionalSample
from bzt.modules.monitoring import LocalClient
from bzt.modules._selenium import SeleniumExecutor
from bzt.modules.services import Unpacker
from bzt.requests_model import has_variable_pattern
from bzt.utils import iteritems, open_browser, BetterDict, ExceptionalDownloader, ProgressBarContext
from bzt.utils import to_json, dehumanize_time, get_full_path, get_files_recursive, replace_in_config
from bzt.modules.blazemeter.blazemeter_reporter import BlazeMeterUploader
from bzt.modules.blazemeter.cloud_test import FUNC_API_TEST_TYPE, FUNC_GUI_TEST_TYPE, TAURUS_TEST_TYPE
from bzt.modules.blazemeter.project_finder import ProjectFinder
from bzt.modules.blazemeter.const import DEDICATED_IPS


CLOUD_CONFIG_BLACK_LIST = {
    "settings": {
        "artifacts-dir": True,
        "aggregator": True,
        "proxy": True,
        "check-updates": True
    },
    "included-configs": True,
    "cli": True,
    "cli-aliases": True,
    "install-id": True,
    "version": True,
    "modules": {
        "jmeter": {
            "path": True,
            "protocol-handlers": True
        },
        "ab": {
            "path": True
        },
        "gatling": {
            "path": True
        },
        "junit": {
            "path": True
        },
        "molotov": {
            "path": True
        },
        "siege": {
            "path": True
        },
        "testng": {
            "path": True
        },
        "tsung": {
            "path": True
        },
        "console": {
            "disable": True,
        },
        "blazemeter": {
            "address": True,
            "data-address": True,
            "token": True,
        },
        "cloud": {
            "address": True,
            "data-address": True,
            "token": True,
        },
    },
    "provisioning": True,

}


class MasterProvisioning(Provisioning):
    def get_rfiles(self):
        rfiles = []
        additional_files = []
        for executor in self.executors:
            executor_rfiles = executor.get_resource_files()
            config = to_json(self.engine.config.get('execution'))
            config += to_json(self.engine.config.get('scenarios'))
            config += to_json(executor.settings)
            for rfile in executor_rfiles:
                if has_variable_pattern(rfile):
                    continue

                if not os.path.exists(self.engine.find_file(rfile)):
                    raise TaurusConfigError("%s: resource file '%s' not found" % (executor, rfile))
                if to_json(rfile) not in config:
                    additional_files.append(rfile)
            rfiles += executor_rfiles

        if additional_files:
            raise TaurusConfigError("Following files can't be handled in cloud: %s" % additional_files)

        rfiles = list(set(rfiles))
        rfiles = [x for x in rfiles if not has_variable_pattern(x)]
        self.log.debug("All resource files are: %s", rfiles)
        return rfiles

    def _fix_filenames(self, old_names):
        # check for concurrent base names
        old_full_names = [self.engine.find_file(x) for x in old_names]
        rbases = [os.path.basename(get_full_path(rfile)) for rfile in old_full_names]
        rpaths = [get_full_path(rfile, step_up=1) for rfile in old_full_names]
        while rbases:
            base, path = rbases.pop(), rpaths.pop()
            if base in rbases:
                index = rbases.index(base)
                if path != rpaths[index]:
                    msg = 'Resource "%s" occurs more than one time, rename to avoid data loss'
                    raise TaurusConfigError(msg % base)

        old_full_names = self.__pack_dirs(old_full_names)
        new_base_names = [os.path.basename(f) for f in old_full_names]
        self.log.debug('Replace file names in config: %s with %s', old_names, new_base_names)
        replace_in_config(self.engine.config, old_names, new_base_names, log=self.log)
        old_full_names = list(set(old_full_names))
        return old_full_names

    def __pack_dirs(self, source_list):
        result_list = []  # files for upload
        packed_list = []  # files for unpacking

        for source in source_list:
            source = get_full_path(source)
            if os.path.isfile(source):
                result_list.append(source)
            else:  # source is dir
                self.log.debug("Compress directory '%s'", source)
                base_dir_name = os.path.basename(source)
                zip_name = self.engine.create_artifact(base_dir_name, '.zip')
                relative_prefix_len = len(os.path.dirname(source))
                with zipfile.ZipFile(zip_name, 'w') as zip_file:
                    for _file in get_files_recursive(source):
                        zip_file.write(_file, _file[relative_prefix_len:])
                result_list.append(zip_name)
                packed_list.append(base_dir_name + '.zip')

        if packed_list:
            services = self.engine.config.get(Service.SERV, [], force_set=True)
            unpacker = BetterDict.from_dict({'module': Unpacker.UNPACK, Unpacker.FILES: packed_list, 'run-at': 'local'})
            services.append(unpacker)

        return result_list


class CloudProvisioning(MasterProvisioning):
    """
    :type user: bzt.bza.User
    :type router: BaseCloudTest
    :type _workspaces: bzt.bza.BZAObjectsList[bzt.bza.Workspace]
    """

    def __init__(self):
        super(CloudProvisioning, self).__init__()
        self.results_url = None
        self.results_reader = None
        self.user = User()
        self.__last_master_status = None
        self.browser_open = 'start'
        self.widget = None
        self.detach = False
        self.router = None
        self.test_ended = False
        self.check_interval = 5.0
        self._last_check_time = None
        self.public_report = False
        self.report_name = None
        self._workspaces = []
        self.launch_existing_test = None
        self.disallow_empty_execution = False

    @staticmethod
    def merge_with_blazemeter_config(module):
        if 'blazemeter' not in module.engine.config.get('modules'):
            module.log.debug("Module 'blazemeter' wasn't found in base config")
            return
        bm_mod = module.engine.instantiate_module('blazemeter')
        bm_settings = copy.deepcopy(bm_mod.settings)
        bm_settings.update(module.settings)
        module.settings = bm_settings

    def _validate_taurus_yml(self):
        if self.router._test.started_taurus_yml_validation():
            self.log.info("Please keep in mind that validation can take time.")
            timeout = 100
            for i in range(timeout):
                if self.router._test.get_taurus_yml_validation():
                    return
                self.log.debug(f"Unsuccessful validation attempt [{i + 1}]. Retrying...")
                sleep(1)
            self.log.error("Unable get validation results!")
        else:
            self.log.error("Unable to validate configuration!")

    def prepare(self):
        CloudProvisioning.merge_with_blazemeter_config(self)
        CloudProvisioning.configure_client(self)
        self._workspaces = self.user.accounts().workspaces()
        if not self._workspaces:
            raise TaurusNetworkError("Your account has no active workspaces, please contact BlazeMeter support")

        self.__dump_locations_if_needed()

        super(CloudProvisioning, self).prepare()
        self.browser_open = self.settings.get("browser-open", self.browser_open)
        self.detach = self.settings.get("detach", self.detach)
        self.check_interval = dehumanize_time(self.settings.get("check-interval", self.check_interval))
        self.public_report = self.settings.get("public-report", self.public_report)
        is_execution_empty = not self.engine.config.get("execution")
        self.launch_existing_test = self.settings.get("launch-existing-test", is_execution_empty, force_set=True)

        finder = ProjectFinder(self.parameters, self.settings, self.user, self._workspaces, self.log)
        finder.default_test_name = "Taurus Cloud Test"

        test_type = self.settings.get("test-type")  # user test type. should we mention it in doc?
        if not test_type:
            func_mode = self.engine.is_functional_mode()
            gui_mode = func_mode and (
                    (len(self.executors) == 1) and
                    isinstance(self.executors[0], SeleniumExecutor))

            if func_mode:
                if gui_mode:
                    test_type = FUNC_GUI_TEST_TYPE
                else:
                    test_type = FUNC_API_TEST_TYPE
            else:
                test_type = TAURUS_TEST_TYPE

        finder.test_type = test_type

        self.router = finder.get_test_router()

        if not self.launch_existing_test:
            self.router.prepare_locations(self.executors, self.engine.config)

            res_files = self.get_rfiles()
            files_for_cloud = self._fix_filenames(res_files)
            self._filter_reporting("blazemeter")    # must not be sent to cloud
            config_for_cloud = self.prepare_cloud_config()
            config_for_cloud.dump(self.engine.create_artifact("cloud", ""))
            del_files = self.settings.get("delete-test-files", True)
            self.router.resolve_test(config_for_cloud, files_for_cloud, del_files)

            self.router.sanitize_test()
            self._validate_taurus_yml()

        self.report_name = self.settings.get("report-name", self.report_name)
        if self.report_name == 'ask' and sys.stdin.isatty():
            self.report_name = input("Please enter report-name: ")

        self.widget = self.get_widget()
        self._filter_reporting("passfail")  # must be sent to claud for conversion

        if self.engine.is_functional_mode():
            self.results_reader = FunctionalBZAReader(self.log)
            self.engine.aggregator.add_underling(self.results_reader)
        else:
            self.results_reader = ResultsFromBZA()
            self.results_reader.log = self.log
            self.engine.aggregator.add_underling(self.results_reader)

    @staticmethod
    def _get_other_modules(config):
        used_classes = LocalClient.__name__, BlazeMeterUploader.__name__
        used_modules = []

        for module in config.get("modules"):
            class_name = config.get("modules").get(module).get("class")
            if class_name and (class_name.split('.')[-1] in used_classes):
                used_modules.append(module)
        return used_modules

    def _get_executors(self):
        executors = []
        for executor in self.executors:
            executors.append(executor.execution.get("executor"))
            if isinstance(executor, SeleniumExecutor):
                executors.append(executor.runner.execution.get("executor"))

        return executors

    def _filter_unused_modules(self, config, provisioning):
        services = [service.get("module") for service in config.get(Service.SERV)]
        reporters = [reporter.get("module") for reporter in config.get(Reporter.REP)]
        consolidator = config.get(SETTINGS).get("aggregator")

        used_modules = self._get_executors() + self._get_other_modules(config)
        used_modules += services + reporters + [consolidator, provisioning]

        modules = set(config.get("modules").keys())
        for module in modules:
            if config.get("modules")[module].get("send-to-blazemeter"):
                continue
            if module not in used_modules:
                del config.get("modules")[module]
            elif config.get("modules")[module].get("class"):
                del config.get("modules")[module]["class"]

    def prepare_cloud_config(self):
        # expand concurrency and throughput
        for executor in self.executors:
            executor.get_load()

        config = copy.deepcopy(self.engine.config)

        provisioning = config.get(Provisioning.PROV)
        self._filter_unused_modules(config, provisioning)

        config['local-bzt-version'] = config.get('version', 'N/A')

        config.filter(CLOUD_CONFIG_BLACK_LIST, black_list=True)

        for execution in config[EXEC]:
            if execution.get("files") == []:
                del execution["files"]

            for param in (ScenarioExecutor.CONCURR, ScenarioExecutor.THRPT):
                param_value = execution.get(param).get(provisioning, None)
                if param_value is None:
                    del execution[param]
                else:
                    execution[param] = param_value

        if self.router.dedicated_ips:
            config[DEDICATED_IPS] = True

        assert isinstance(config, Configuration)
        return config

    def __dump_locations_if_needed(self):
        if self.settings.get("dump-locations", False):
            locations = {}
            for loc in self._workspaces.locations(include_private=True):
                locations[loc['id']] = loc

            data = [("ID", "Name")]
            for location_id in sorted(locations):
                location = locations[location_id]
                data.append((location_id, location['title']))
            table = SingleTable(data) if sys.stdout and sys.stdout.isatty() else AsciiTable(data)
            self.log.warning("Dumping available locations instead of running the test:\n%s", table.table)
            raise NormalShutdown("Done listing locations")

    def _filter_reporting(self, mod_name):
        reporting = self.engine.config.get(Reporter.REP, [])
        new_reporting = []
        for index, reporter in enumerate(reporting):
            exc = TaurusConfigError("'module' attribute not found in %s" % reporter)
            cls = reporter.get('module', exc)
            if cls == mod_name:
                self.log.debug(f"Remove {mod_name} for Cloud provisioning: {reporter}")
            else:
                new_reporting.append(reporter)

        self.engine.config[Reporter.REP] = new_reporting

    @staticmethod
    def configure_client(module):
        module.user.log = module.log
        module.user.logger_limit = module.settings.get("request-logging-limit", module.user.logger_limit)
        module.user.address = module.settings.get("address", module.user.address)
        module.user.token = module.settings.get("token", module.user.token)
        module.user.timeout = dehumanize_time(module.settings.get("timeout", module.user.timeout))
        if isinstance(module.user.http_session, requests.Session):
            module.log.debug("Installing http client")
            module.user.http_session = module.engine.get_http_client()
            module.user.http_request = module.user.http_session.request
        if not module.user.token:
            raise TaurusConfigError("You must provide API token to use cloud provisioning")

    def startup(self):
        super(CloudProvisioning, self).startup()
        self.results_url = self.router.launch_test()
        self.log.info("Started cloud test: %s", self.results_url)
        if self.results_url:
            if self.browser_open in ('start', 'both'):
                open_browser(self.results_url)

        if self.user.token and self.public_report:
            public_link = self.router.master.make_report_public()
            self.log.info("Public report link: %s", public_link)

        if self.report_name:
            self.router.master.set({"name": str(self.report_name)})

    def _should_skip_check(self):
        now = time.time()
        if self._last_check_time is None:
            return False
        elif now >= self._last_check_time + self.check_interval:
            return False
        else:
            return True

    def check(self):
        if self.detach:
            self.log.warning('Detaching Taurus from started test...')
            return True

        if self._should_skip_check():
            self.log.debug("Skipping cloud status check")
            return False

        self._last_check_time = time.time()

        master = self._check_master_status()
        status = master.get('status')
        progress = master.get('progress')   # number value of status, see BZA API

        if status != self.__last_master_status:
            self.__last_master_status = status
            self.log.info("Cloud test status: %s", status)

        if self.results_reader and progress and progress >= BZA_TEST_DATA_RECEIVED:
            self.results_reader.master = self.router.master

        if progress == ENDED:
            self.log.info("Test was stopped in the cloud: %s", status)
            self.test_ended = True
            return True

        self.router.start_if_ready()

        self.widget.update()
        return super(CloudProvisioning, self).check()

    def _check_master_status(self):
        return self.router.get_master_status()

    def post_process(self):
        self.log.warning('Part of result data might be missed here due to BM API specifics')

        if not self.detach and self.router and not self.test_ended:
            self.router.stop_test()

        if self.results_url:
            if self.browser_open in ('end', 'both'):
                open_browser(self.results_url)

        if self.router and self.router.master:
            full = self.router.master.get_full()
            if 'note' in full and full['note']:
                self.log.warning("Cloud test has probably failed with message: %s", full['note'])

            for session in full.get('sessions', ()):
                for error in session.get("errors", ()):
                    raise TaurusException(to_json(error))

            if "hasThresholds" in full and full["hasThresholds"]:
                thresholds = self.router.master.get_thresholds()
                for item in thresholds.get('data', []):
                    if item.get('success', None) is False:
                        reason = None
                        for assertion in item.get('assertions', []):
                            if assertion.get('success', None) is False:
                                criterion = assertion.get('field', '')
                                label = assertion.get('label', '')
                                reason = "Cloud failure criterion %r (on label %r) was met" % (criterion, label)
                                break
                        if reason is None:
                            reason = "Cloud tests failed because failure criteria were met"
                        self.log.warning(reason)
                        raise AutomatedShutdown(reason)

            # if we have captured HARs, let's download them
            for service in self.engine.config.get(Service.SERV, []):
                mod = service.get('module', TaurusConfigError("No 'module' specified for service"))
                assert isinstance(mod, str), mod
                module = self.engine.instantiate_module(mod)
                if isinstance(module, ServiceStubCaptureHAR):
                    self._download_logs()
                    break

            if "functionalSummary" in full:
                summary = full["functionalSummary"]
                if summary is None or summary.get("isFailed", False):
                    raise AutomatedShutdown("Cloud tests failed")

    def _download_logs(self):
        for session in self.router.master.sessions():
            assert isinstance(session, Session)
            for log in session.get_logs():
                self.log.info("Downloading %s from the cloud", log['filename'])
                cloud_dir = os.path.join(self.engine.artifacts_dir, 'cloud-artifacts')
                if not os.path.exists(cloud_dir):
                    os.makedirs(cloud_dir)
                dest = os.path.join(cloud_dir, log['filename'])
                dwn = ExceptionalDownloader(self.engine.get_http_client())
                with ProgressBarContext() as pbar:
                    try:
                        dwn.get(log['dataUrl'], dest, reporthook=pbar.download_callback)
                    except BaseException:
                        self.log.debug("Error is: %s", traceback.format_exc())
                        self.log.warning("Failed to download from %s", log['dataUrl'])
                        continue

                    if log['filename'].startswith('artifacts') and log['filename'].endswith('.zip'):
                        with zipfile.ZipFile(dest) as zipf:
                            for name in zipf.namelist():
                                ext = name.split('.')[-1].lower()
                                if ext in ('har', 'jpg', 'js', 'html', 'css'):
                                    self.log.debug("Extracting %s to %s", name, cloud_dir)
                                    zipf.extract(name, cloud_dir)

    def get_widget(self):
        if not self.widget:
            self.widget = CloudProvWidget(self.router)
        return self.widget


class ResultsFromBZA(ResultsProvider):
    """
    :type master: bzt.bza.Master
    """

    def __init__(self, master=None):
        super(ResultsFromBZA, self).__init__()
        self.master = master
        self.min_ts = 0
        self.log = logging.getLogger('')
        self.prev_errors = BetterDict()
        self.cur_errors = BetterDict()
        self.handle_errors = True

    def _get_err_diff(self):
        # find diff of self.prev_errors and self.cur_errors
        diff = {}
        for label in self.cur_errors:
            if label not in self.prev_errors:
                diff[label] = self.cur_errors[label]
                continue

            for msg in self.cur_errors[label]:
                if msg not in self.prev_errors[label]:
                    prev_count = 0
                else:
                    prev_count = self.prev_errors[label][msg]['count']

                delta = self.cur_errors[label][msg]['count'] - prev_count
                if delta > 0:
                    if label not in diff:
                        diff[label] = {}
                    diff[label][msg] = {'count': delta, 'rc': self.cur_errors[label][msg]['rc']}

        return {k: diff[k] for k in diff if diff[k]}  # clean from empty items

    def _calculate_datapoints(self, final_pass=False):
        if self.master is None:
            return

        data, aggr_raw = self.query_data()
        aggr = {}
        for label in aggr_raw:
            aggr[label['labelName']] = label

        for label in data:
            if label.get('kpis') and not final_pass:
                label['kpis'].pop(-1)  # never take last second since it could be incomplete

        timestamps = []
        for label in data:
            if label.get('label') == 'ALL':
                timestamps.extend([kpi['ts'] for kpi in label.get('kpis', [])])

        self.handle_errors = True

        for tstmp in timestamps:
            point = DataPoint(tstmp)
            point[DataPoint.SOURCE_ID] = self.master['id']
            self.__generate_kpisets(aggr, data, point, tstmp)

            if self.handle_errors:
                self.handle_errors = False
                self.cur_errors = self.__get_errors_from_bza()
                err_diff = self._get_err_diff()
                if err_diff:
                    self.__add_err_diff(point, err_diff)
                    self.prev_errors = self.cur_errors

            point.recalculate()

            self.min_ts = point[DataPoint.TIMESTAMP] + 1
            yield point

    def __add_err_diff(self, point, err_diff):
        for label in err_diff:
            point_label = '' if label == 'ALL' else label
            if point_label not in point[DataPoint.CURRENT]:
                self.log.warning("Got inconsistent kpi/error data for label: %s", point_label)
                kpiset = KPISet()
                point[DataPoint.CURRENT][point_label] = kpiset
                kpiset[KPISet.SAMPLE_COUNT] = sum([item['count'] for item in err_diff[label].values()])
            else:
                kpiset = point[DataPoint.CURRENT][point_label]

            kpiset[KPISet.ERRORS] = self.__get_kpi_errors(err_diff[label])
            kpiset[KPISet.FAILURES] = sum([x['cnt'] for x in kpiset[KPISet.ERRORS]])
            kpiset[KPISet.SAMPLE_COUNT] = kpiset[KPISet.SUCCESSES] + kpiset[KPISet.FAILURES]
            assert kpiset[KPISet.SAMPLE_COUNT] > 0, point_label

    def __generate_kpisets(self, aggr, data, point, tstmp):
        for label in data:
            for kpi in label.get('kpis', []):
                if kpi['ts'] != tstmp:
                    continue
                label_str = label.get('label')
                if label_str is None or label_str not in aggr:
                    self.log.warning("Skipping inconsistent data from API for label: %s", label_str)
                    continue

                if kpi['n'] <= 0:
                    self.log.warning("Skipping empty KPI item got from API: %s", kpi)
                    continue

                kpiset = self.__get_kpiset(aggr, kpi, label_str)
                point[DataPoint.CURRENT]['' if label_str == 'ALL' else label_str] = kpiset

    def __get_errors_from_bza(self):
        #
        # This method reads error report from BZA
        #
        # internal errors format:
        # <request_label>:
        #   <error_message>:
        #     'count': <count of errors>
        #     'rc': <response code>
        #
        result = {}
        try:
            errors = self.master.get_errors()
        except (URLError, TaurusNetworkError):
            self.log.warning("Failed to get errors, will retry in %s seconds...", self.master.timeout)
            self.log.debug("Full exception: %s", traceback.format_exc())
            time.sleep(self.master.timeout)
            errors = self.master.get_errors()
            self.log.info("Succeeded with retry")

        for e_record in errors:
            _id = e_record["_id"]
            if _id == "ALL":
                _id = ""
            result[_id] = {}
            for error in e_record['errors']:
                result[_id][error['m']] = {'count': error['count'], 'rc': error['rc']}
            for assertion in e_record['assertions']:
                result[_id][assertion['failureMessage']] = {'count': assertion['failures'], 'rc': assertion['name']}
        return result

    def __get_kpi_errors(self, errors):
        result = []
        for msg in errors:
            kpi_error = KPISet.error_item_skel(
                error=msg,
                ret_c=errors[msg]['rc'],
                cnt=errors[msg]['count'],
                err_type=KPISet.ERRTYPE_ERROR,
                urls=Counter(), tag=None)
            result.append(kpi_error)
        return result

    def __get_kpiset(self, aggr, kpi, label):
        kpiset = KPISet()
        kpiset[KPISet.FAILURES] = kpi['ec']
        kpiset[KPISet.CONCURRENCY] = kpi['na']
        kpiset[KPISet.SAMPLE_COUNT] = kpi['n']
        assert kpi['n'] > 0 and kpi['n'] >= kpi['ec']
        kpiset[KPISet.SUCCESSES] = kpi['n'] - kpi['ec']
        kpiset.sum_rt += kpi['t_avg'] * kpi['n'] / 1000.0
        kpiset.sum_lt += kpi['lt_avg'] * kpi['n'] / 1000.0
        perc_map = {'90line': 90.0, "95line": 95.0, "99line": 99.0}
        for field, level in iteritems(perc_map):
            kpiset[KPISet.PERCENTILES][str(level)] = aggr[label][field] / 1000.0
        return kpiset

    def query_data(self):
        try:
            data = self.master.get_kpis(self.min_ts)
        except (URLError, TaurusNetworkError):
            self.log.warning("Failed to get result KPIs, will retry in %s seconds...", self.master.timeout)
            self.log.debug("Full exception: %s", traceback.format_exc())
            time.sleep(self.master.timeout)
            data = self.master.get_kpis(self.min_ts)
            self.log.info("Succeeded with retry")

        try:
            aggr = self.master.get_aggregate_report()
        except (URLError, TaurusNetworkError):
            self.log.warning("Failed to get aggregate results, will retry in %s seconds...", self.master.timeout)
            self.log.debug("Full exception: %s", traceback.format_exc())
            time.sleep(self.master.timeout)
            aggr = self.master.get_aggregate_report()
            self.log.info("Succeeded with retry")

        return data, aggr


class FunctionalBZAReader(FunctionalResultsReader):
    def __init__(self, parent_log, master=None):
        super(FunctionalBZAReader, self).__init__()
        self.master = master
        self.log = parent_log.getChild(self.__class__.__name__)

    @staticmethod
    def extract_samples_from_group(group, group_summary):
        group_name = group_summary.get("name") or "Tests"
        for sample in group["samples"]:
            status = "PASSED"
            if sample["error"]:
                status = "FAILED"
            error_msg = ""
            error_trace = ""
            assertions = sample.get("assertions")
            if assertions:
                for assertion in assertions:
                    if assertion.get("isFailed"):
                        error_msg = assertion.get("errorMessage")
                        status = "BROKEN"

            rtm = sample.get("responseTime") or 0.0
            yield FunctionalSample(
                test_case=sample["label"],
                test_suite=group_name,
                status=status,
                start_time=int(sample["created"]),
                duration=rtm / 1000.0,
                error_msg=error_msg,
                error_trace=error_trace,
                extras={},
                subsamples=[],
            )

    def read(self, last_pass=False):
        if self.master is None:
            return

        if last_pass:
            try:
                groups = self.master.get_functional_report_groups()
            except (URLError, TaurusNetworkError):
                self.log.warning("Failed to get test groups, will retry in %s seconds...", self.master.timeout)
                self.log.debug("Full exception: %s", traceback.format_exc())
                time.sleep(self.master.timeout)
                groups = self.master.get_functional_report_groups()
                self.log.info("Succeeded with retry")

            for group_summary in groups:
                group_id = group_summary['groupId']
                try:
                    group = self.master.get_functional_report_group(group_id)
                except (URLError, TaurusNetworkError):
                    self.log.warning("Failed to get test group, will retry in %s seconds...", self.master.timeout)
                    self.log.debug("Full exception: %s", traceback.format_exc())
                    time.sleep(self.master.timeout)
                    group = self.master.get_functional_report_group(group_id)
                    self.log.info("Succeeded with retry")

                for sample in self.extract_samples_from_group(group, group_summary):
                    yield sample


class CloudProvWidget(Pile, PrioritizedWidget):
    def __init__(self, test):
        """
        :type test: BaseCloudTest
        """
        self.test = test
        self.text = Text("")
        super(CloudProvWidget, self).__init__([self.text])
        PrioritizedWidget.__init__(self)

    def update(self):
        txt = self.test.get_test_status_text()
        if txt:
            self.text.set_text(txt)


class ServiceStubScreenshoter(Service):
    def startup(self):
        if not isinstance(self.engine.provisioning, CloudProvisioning):
            self.log.warning("Stub for service 'screenshoter', use cloud provisioning to have it working")


class ServiceStubCaptureHAR(Service):
    def startup(self):
        if not isinstance(self.engine.provisioning, CloudProvisioning):
            self.log.warning("Stub for service 'capturehar', use cloud provisioning to have it working")



"""
Copyright 2018 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import sys

from bzt import TaurusConfigError
from bzt.engine import SETTINGS
from bzt.modules import SubprocessedExecutor, ConsolidatingAggregator, FuncSamplesReader
from bzt.modules.functional import FunctionalResultsReader
from bzt.modules.jmeter import JTLReader
from bzt.utils import FileReader, get_full_path, BZT_DIR, get_assembled_value, shutdown_process
from .generator import ApiritifScriptGenerator
from .._selenium import Selenium
from ..services import PythonTool

IGNORED_LINE = re.compile(r"[^,]+,Total:\d+ Passed:\d+ Failed:\d+")


class ApiritifNoseExecutor(SubprocessedExecutor):
    """
    :type _tailer: FileReader
    """

    def __init__(self):
        super(ApiritifNoseExecutor, self).__init__()
        self._tailer = FileReader(file_opener=lambda _: None, parent_logger=self.log)
        self.apiritif = None
        self.selenium = None
        self.test_mode = None

    def resource_files(self):
        files = super(ApiritifNoseExecutor, self).resource_files()
        for source in self.get_scenario().get_data_sources():
            files.append(source['path'])

        return files

    def create_func_reader(self, report_file):
        del report_file
        return ApiritifFuncReader(self.engine, self.log)

    def create_load_reader(self, report_file):
        del report_file
        reader = ApiritifLoadReader(self.log)
        reader.engine = self.engine
        return reader

    def prepare(self):
        super(ApiritifNoseExecutor, self).prepare()
        self.install_required_tools()
        self.script = self.get_script_path()

        self.test_mode = self.execution.get("test-mode", "apiritif")
        if self.test_mode != "apiritif":
            self.selenium = self._get_tool(Selenium, engine=self.engine, settings=self.settings)
            if not self.selenium.check_if_installed():
                self.selenium.install()

        if not self.script:
            if "requests" in self.get_scenario():
                self.script = self.__tests_from_requests()
            else:
                raise TaurusConfigError("Nothing to test, no requests were provided in scenario")

        # path to taurus dir. It's necessary for bzt usage inside tools/helpers
        self.env.add_path({"PYTHONPATH": get_full_path(BZT_DIR, step_up=1)})

        if self.settings.get("plugins-path"):
            # add path to plugins directory to Apiritif env vars
            self.log.debug(f'Found Apiritif plugins path: {self.settings.get("plugins-path")}')
            self.env.add_path({"PLUGINS_PATH": self.settings.get('plugins-path')})

        self.reporting_setup()  # no prefix/suffix because we don't fully control report file names

    def __tests_from_requests(self):
        filename = self.engine.create_artifact("test_requests", ".py")
        scenario = self.get_scenario()

        bzm_tdo_settings = None
        if self.engine.config.get("settings").get("master_publish_url"):
            bzm_tdo_settings = {
                "master_publish_url": self.engine.config.get("settings").get("master_publish_url")
            }

        if self.test_mode == "apiritif":
            builder = ApiritifScriptGenerator(scenario, self.label, executor=self, test_mode=self.test_mode,
                                              bzm_tdo_settings=bzm_tdo_settings)
            builder.verbose = self.__is_verbose()
        else:
            wdlog = self.engine.create_artifact('webdriver', '.log')

            generate_markers = self.settings.get('generate-flow-markers', None)
            generate_markers = scenario.get('generate-flow-markers', generate_markers)

            scenario_caps = scenario.get("capabilities")

            if isinstance(scenario_caps, list):
                self.log.warning("Obsolete format of capabilities found (list), should be dict")
                scenario["capabilities"] = {item.keys()[0]: item.values()[0] for item in scenario_caps}

            configs = (self.settings, scenario, self.execution)

            capabilities = get_assembled_value(configs, "capabilities")
            remote = get_assembled_value(configs, "remote")

            selenium_version = self.selenium.get_version()

            builder = ApiritifScriptGenerator(
                scenario, self.label, wdlog, executor=self,
                ignore_unknown_actions=self.settings.get("ignore-unknown-actions", False),
                generate_markers=generate_markers,
                capabilities=capabilities,
                wd_addr=remote, test_mode=self.test_mode,
                generate_external_handler=True if self.settings.get('plugins-path', False) else False,
                selenium_version=selenium_version,
                bzm_tdo_settings=bzm_tdo_settings)

        builder.build_source_code()
        builder.save(filename)
        if isinstance(self.engine.aggregator, ConsolidatingAggregator) and isinstance(builder, ApiritifScriptGenerator):
            self.engine.aggregator.ignored_labels.extend(builder.service_methods)
        return filename

    def get_load(self):
        load = super().get_load()
        # should use raw iterations
        iterations = self.get_raw_load().iterations
        if self.get_raw_load().iterations is None:
            iterations = self.__calculate_iterations(load)
        iterations = self.__get_iterations_from_test_data(iterations)
        self.log.info("Iterations set to %s", iterations)
        return self.LOAD_FMT(concurrency=load.concurrency, ramp_up=load.ramp_up, throughput=load.throughput,
                             hold=load.hold, iterations=iterations, duration=load.duration, steps=load.steps)

    def __get_iterations_from_test_data(self, default_iterations):
        data_iteration = self.engine.config.get("dependencies").get("data").get("iteration").get("repeat")
        iterations = default_iterations
        if data_iteration:
            iterations = data_iteration
            self.log.debug("Iterations limit overridden by test data setting, set to %s", iterations)
        return iterations

    def __calculate_iterations(self, load):
        msg = "No iterations limit in config, choosing anything... set "
        if load.duration or self.engine.is_functional_mode() and list(self.get_scenario().get_data_sources()):
            iterations = 0  # infinite for func mode and ds
            msg += "0 (infinite) as "
            if load.duration:
                msg += "duration found (hold-for + ramp-up)"
            elif self.engine.is_functional_mode():
                msg += "taurus works in functional mode"
            else:
                msg += "data-sources found"
        else:
            iterations = 1  # run once otherwise
            msg += "1"

        self.log.debug(msg)

        return iterations

    def install_required_tools(self):
        self.apiritif = self._get_tool(Apiritif, engine=self.engine, settings=self.settings)
        if not self.apiritif.check_if_installed():
            self.apiritif.install()

    def startup(self):
        executable = self.settings.get("interpreter", sys.executable)

        report_type = ".ldjson" if self.engine.is_functional_mode() else ".csv"
        report_tpl = self.engine.create_artifact("apiritif", ".") + "%s" + report_type
        cmdline = [executable, "-m", "apiritif.loadgen", '--result-file-template', report_tpl]

        load = self.get_load()
        if load.concurrency:
            cmdline += ['--concurrency', str(load.concurrency)]

        if load.iterations:
            cmdline += ['--iterations', str(load.iterations)]

        if load.hold:
            cmdline += ['--hold-for', str(load.hold)]

        if load.ramp_up:
            cmdline += ['--ramp-up', str(load.ramp_up)]

        if load.steps:
            cmdline += ['--steps', str(load.steps)]

        if self.__is_verbose():
            cmdline += ['--verbose']

        cmdline += [self.script]
        self.process = self._execute(cmdline)
        self._tailer = FileReader(filename=self.stdout.name, parent_logger=self.log)

    def has_results(self):
        if not self.reader:
            return False
        return self.reader.read_records

    @staticmethod
    def _normalize_label(label):
        for char in ":/":
            if char in label:
                label = label.replace(char, '_')
        return label

    def _check_stdout(self):
        for line in self._tailer.get_lines():
            if "Adding worker" in line:
                marker = "results="
                pos = line.index(marker)
                fname = line[pos + len(marker):].strip()
                self.log.debug("Adding result reader for %s", fname)
                self.reader.register_file(fname)
            elif "Transaction started" in line:
                colon = line.index('::')
                values = {
                    part.split('=')[0]: part.split('=')[1]
                    for part in line[colon + 2:].strip().split(',')
                }
                label = self._normalize_label(values['name'])
                start_time = float(values['start_time'])
                self.transaction_started(label, start_time)
            elif "Transaction ended" in line:
                colon = line.index('::')
                values = {
                    part.split('=')[0]: part.split('=')[1]
                    for part in line[colon + 2:].strip().split(',')
                }
                label = self._normalize_label(values['name'])
                duration = float(values['duration'])
                self.transacion_ended(label, duration)

    def check(self):
        self._check_stdout()
        return super(ApiritifNoseExecutor, self).check()

    def __log_lines(self):
        lines = []
        for line in self._tailer.get_lines():
            if not IGNORED_LINE.match(line):
                lines.append(line)

        if lines:
            self.log.info("\n".join(lines))

    def shutdown(self):
        shutdown_process(self.process, self.log, send_sigterm=False)

    def post_process(self):
        self._check_stdout()
        self.__log_lines()
        self._tailer.close()
        self.apiritif.post_process()
        if self.selenium:
            self.selenium.post_process()
        super(ApiritifNoseExecutor, self).post_process()

    def __is_verbose(self):
        engine_verbose = self.engine.config.get(SETTINGS).get("verbose", False)
        executor_verbose = self.settings.get("verbose", engine_verbose)
        return executor_verbose


class ApiritifTester(ApiritifNoseExecutor):
    pass


class Apiritif(PythonTool):
    VERSION = "1.1.3"
    PACKAGES = ["apiritif"]


class ApiritifLoadReader(ConsolidatingAggregator):
    def __init__(self, parent_log):
        super(ApiritifLoadReader, self).__init__()
        self.log = parent_log.getChild(self.__class__.__name__)

    def register_file(self, report_filename):
        reader = JTLReader(report_filename, self.log)
        self.add_underling(reader)

    @property
    def read_records(self):
        for reader in self.underlings:  # type: JTLReader
            if reader.read_records > 0:
                return True
        return False


class ApiritifFuncReader(FunctionalResultsReader):
    def __init__(self, engine, parent_log):
        super(ApiritifFuncReader, self).__init__()
        self.engine = engine
        self.log = parent_log.getChild(self.__class__.__name__)
        self.filenames = []
        self.readers = []
        self.read_records = False

    def register_file(self, report_filename):
        self.filenames.append(report_filename)
        reader = FuncSamplesReader(report_filename, self.engine, self.log)
        self.readers.append(reader)

    def read(self, last_pass=False):
        for reader in self.readers:
            for sample in reader.read(last_pass):
                if not self.read_records:
                    self.read_records = True
                yield sample



"""
Modern Retry Decorator for BZA API Classes

This module provides annotation-based retry logic that:
- Retries on IOError, URLError, SSLError, ReadTimeout
- Retries on HTTP 5xx errors (server errors)
- Retries on HTTP 429 (rate limiting)
- Does NOT retry on 4xx errors (client errors)
- Supports exponential backoff with optional jitter
- Keeps transient methods (send_kpi_data, upload_file) silent on final failure
- Logs all retry attempts
"""

import functools
import logging
import time
from typing import Callable, Optional, Tuple, Type

from requests.exceptions import ReadTimeout, RequestException
from urllib.error import URLError
from ssl import SSLError
from urllib3.exceptions import ProtocolError, ReadTimeoutError

from bzt import TaurusNetworkError

LOG = logging.getLogger("BZA.Retry")

# Network problems that should trigger retry
NETWORK_PROBLEMS = (IOError, URLError, SSLError, ReadTimeout, TaurusNetworkError, ProtocolError, ReadTimeoutError)

# HTTP status codes that should trigger retry
RETRIABLE_STATUS_CODES = (429, 500, 501, 502, 503, 504)

# HTTP status codes that should NOT retry
NON_RETRIABLE_STATUS_CODES = (400, 401, 403, 404, 405, 406, 409, 410)


class RetryConfig:
    """Configuration for retry behavior"""

    def __init__(
        self,
        max_attempts: int = 3,
        initial_delay: float = 1.0,
        backoff_factor: float = 2.0,
        max_delay: float = 60.0,
        jitter: bool = False,
        transient: bool = False,
        retriable_exceptions: Tuple[Type[Exception], ...] = NETWORK_PROBLEMS,
    ):
        self.max_attempts = max_attempts
        self.initial_delay = initial_delay
        self.backoff_factor = backoff_factor
        self.max_delay = max_delay
        self.jitter = jitter
        self.transient = transient
        self.retriable_exceptions = retriable_exceptions

    def calculate_delay(self, attempt: int) -> float:
        """Calculate delay for given attempt number (0-indexed)"""
        delay = self.initial_delay * (self.backoff_factor ** attempt)
        delay = min(delay, self.max_delay)

        if self.jitter:
            import random
            jitter_factor = random.uniform(0.75, 1.25)
            delay *= jitter_factor

        return delay


def is_retriable_exception(exc: Exception) -> bool:
    """Check if exception should trigger retry"""
    return isinstance(exc, NETWORK_PROBLEMS)


def is_retriable_status(status_code: int) -> bool:
    """Check if HTTP status code should trigger retry"""
    return status_code in RETRIABLE_STATUS_CODES or (500 <= status_code < 600)


def extract_status_from_error(error: Exception) -> Optional[int]:
    """Extract HTTP status code from exception if present"""
    if hasattr(error, 'status_code'):
        return error.status_code

    if hasattr(error, 'response') and hasattr(error.response, 'status_code'):
        return error.response.status_code

    return None


def should_retry(exc: Exception, config: RetryConfig) -> bool:
    """
    Determine if exception should trigger retry

    Args:
        exc: The exception that occurred
        config: Retry configuration

    Returns:
        True if should retry, False if should fail
    """
    # Check if it's a retriable exception type
    if is_retriable_exception(exc):
        return True

    # Check for HTTP status codes in the error
    status_code = extract_status_from_error(exc)
    if status_code is not None:
        if status_code in NON_RETRIABLE_STATUS_CODES:
            return False
        if is_retriable_status(status_code):
            return True

    # For RequestException, check response status
    if isinstance(exc, RequestException):
        if hasattr(exc, 'response') and exc.response is not None:
            if is_retriable_status(exc.response.status_code):
                return True
            if exc.response.status_code in NON_RETRIABLE_STATUS_CODES:
                return False

    return False


def retry(
    max_attempts: int = 3,
    initial_delay: float = 1.0,
    backoff_factor: float = 2.0,
    max_delay: float = 60.0,
    jitter: bool = False,
    transient: bool = False,
):
    """
    Decorator for retrying methods with exponential backoff

    Args:
        max_attempts: Maximum number of attempts (default 3)
        initial_delay: Initial delay between retries in seconds (default 1.0)
        backoff_factor: Exponential backoff multiplier (default 2.0)
        max_delay: Maximum delay between retries (default 60.0)
        jitter: Add random jitter to delays (default False)
        transient: If True, silently fail on final attempt (default False)
    """
    config = RetryConfig(
        max_attempts=max_attempts,
        initial_delay=initial_delay,
        backoff_factor=backoff_factor,
        max_delay=max_delay,
        jitter=jitter,
        transient=transient,
    )

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            logger = logging.getLogger(f"{self.__class__.__name__}.{func.__name__}")
            last_exception = None

            for attempt in range(config.max_attempts):
                try:
                    if attempt > 0:
                        logger.info(
                            f"Retry attempt {attempt + 1}/{config.max_attempts} for {func.__name__}"
                        )

                    return func(self, *args, **kwargs)

                except (Exception, BaseException) as exc:
                    last_exception = exc

                    # Check if this exception should trigger a retry
                    if not should_retry(exc, config):
                        logger.debug(
                            f"Error in {func.__name__} is not retriable: {exc.__class__.__name__}"
                        )
                        raise

                    # Check if we have more attempts
                    if attempt >= config.max_attempts - 1:
                        logger.warning(
                            f"All {config.max_attempts} attempts failed for {func.__name__}: {exc}"
                        )

                        if config.transient:
                            logger.info(
                                f"Transient operation {func.__name__} failed after {config.max_attempts} attempts, "
                                f"but continuing (transient=True)"
                            )
                            return None
                        else:
                            raise

                    # Calculate delay for next attempt
                    delay = config.calculate_delay(attempt)
                    logger.debug(
                        f"Attempt {attempt + 1} failed with {exc.__class__.__name__}, "
                        f"retrying in {delay:.2f}s: {exc}"
                    )

                    # Sleep before retry
                    time.sleep(delay)

            # Should never reach here, but just in case
            if last_exception:
                raise last_exception

        wrapper._retry_config = config
        return wrapper

    return decorator


# Convenience decorators for common patterns

def retry_once(transient: bool = False):
    """Retry once (2 total attempts) with 1 second delay"""
    return retry(max_attempts=2, initial_delay=1.0, backoff_factor=1.0, transient=transient)


def retry_aggressive(transient: bool = False):
    """Aggressive retry: 5 attempts with exponential backoff and jitter"""
    return retry(
        max_attempts=5,
        initial_delay=0.5,
        backoff_factor=1.5,
        max_delay=60.0,
        jitter=True,
        transient=transient
    )


def retry_gentle(transient: bool = False):
    """Gentle retry: 3 attempts with slow backoff, good for rate limiting"""
    return retry(
        max_attempts=3,
        initial_delay=2.0,
        backoff_factor=2.0,
        max_delay=60.0,
        jitter=True,
        transient=transient
    )


def retry_transient(max_attempts: int = 3):
    """Retry with silent failure (for data submission operations)"""
    return retry(max_attempts=max_attempts, backoff_factor=2.0, transient=True)


def retry_critical(max_attempts: int = 5):
    """Retry with failure (for critical operations that should not silently fail)"""
    return retry(max_attempts=max_attempts, backoff_factor=1.5, jitter=True, transient=False)


"""
Copyright 2024 BlazeMeter Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
import socket
import traceback
import time
from math import floor
from collections import deque
from multiprocessing import AuthenticationError
from multiprocessing.connection import Client
import logging

logging.basicConfig(stream=sys.stdout, level=logging.WARNING,
                    format="%(asctime)s %(levelname)s: %(message)s")

class JmeterRampupProcess(object):
    def __init__(self, beanshells, rampup_addr, rampup_port, rampup_pass, artifacts_dir, log_level=logging.INFO):
        self.beanshells = beanshells
        self.rampup_addr = rampup_addr
        self.rampup_port = rampup_port
        self.rampup_pass = rampup_pass
        self.filename = os.path.join(artifacts_dir, "kpi.jtl")
        self.log = logging.getLogger("Rampup.Process")
        self.log.setLevel(log_level)

    def run(self):
        res_list = None
        cur_goal = None
        while True:
            new_rampup = self._check_change_params()
            if new_rampup:
                self.log.info(f"Got new rampup configuration: {new_rampup}")
                res_list = self._calc_rampup(self.filename, new_rampup)
                cur_goal = res_list.popleft() if res_list else None
            if cur_goal:
                now = self._now()
                do_goal = None
                while cur_goal and cur_goal[1] < now:
                    do_goal = cur_goal
                    cur_goal = res_list.popleft() if res_list else None
                if do_goal:
                    self.log.info(f"Setting concurrency: {do_goal}")
                    self._change_concurrencies(do_goal[0])
            if self._stop():
                break
            time.sleep(0.5)

    def _now(self):
        return time.time()

    def _stop(self):
        return False

    def _socket_recv(self, sock_a):
        read = ""
        try:
            for _ in range(0, 1000):  # just safety valve
                read += sock_a.recv(4096).decode()
        except socket.timeout:
            self.log.debug("Timeout reading from beanshell socket")
        except BaseException:
            self.log.warning("Failed to read response from beanshell server %s: %s" %
                  (sock_a.getpeername(), traceback.format_exc()))

        lines = read.splitlines(True)
        result = []
        for line in lines:
            if line.startswith("BeanShell "):
                continue

            while line.startswith("bsh % "):
                line = line[len("bsh % "):]

            line = line.strip()
            result.append(line)

        return "".join(result)

    def _get_current_concurrency(self, file):
        try:
            with open(file, 'rb') as f:
                try:
                    f.seek(-1, os.SEEK_END)
                    # Find last completed line (ending with \n)
                    while f.read(1) != b'\n':
                        f.seek(-2, os.SEEK_CUR)
                    f.seek(-2, os.SEEK_CUR)
                    while f.read(1) != b'\n':
                        f.seek(-2, os.SEEK_CUR)
                except OSError:
                    f.seek(0)
                last_line = f.readline().decode()
            try:
                cur_con = last_line.split(",")[-5]
            except IndexError:
                self.log.warning("Could not read users retry next loop")
                cur_con = False
        except FileNotFoundError:
            self.log.warning(f"File {self.filename} not found. Retry next loop")
            cur_con = False

        self.log.info(f"Current concurrency: {cur_con}")
        return cur_con

    def _change_concurrencies(self, desired_users):
        for beanshell_addr, beanshell_port in self.beanshells:
            self._change_concurrency(desired_users, beanshell_addr, beanshell_port)

    def _change_concurrency(self, desired_users, beanshell_addr, beanshell_port):
        result = ""
        socket_factory = socket.socket
        sock = socket_factory(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect((beanshell_addr, beanshell_port + 1))
            sock.settimeout(1)
            script = (f'org.apache.jmeter.util.JMeterUtils.setProperty("BM_CTG_RampUp","0");'
                      f'org.apache.jmeter.util.JMeterUtils.setProperty("BM_CTG_TargetLevel","{desired_users}");')
            sock.sendall(script.encode())

            result += self._socket_recv(sock) + "\n"
        except BaseException:
            self.log.warning("Failed to send command to beanshell server %s:%s: %s" %
                  (beanshell_addr, beanshell_port, traceback.format_exc()))
        finally:
            sock.close()
        self.log.debug(f"Beanshell recv: {result}")

    def _check_change_params(self):
        address = (self.rampup_addr, self.rampup_port)

        try:
            with Client(address, authkey=self.rampup_pass.encode()) as conn:
                return conn.recv()
        except ConnectionRefusedError:
            pass
        except AuthenticationError:
            self.log.fatal("Forbidden. Rampup server returns AuthenticationError.")
            print(f"Forbidden. Rampup server returns AuthenticationError: {traceback.format_exc()}",
                  file=sys.stderr, flush=True)
            sys.exit(1)

        return None

    def _calc_rampup(self, file, new_rampup):
        users = new_rampup["concurrency"]
        if not users:
            return None
        rampup = new_rampup["ramp_up_duration"]
        steps = new_rampup["ramp_up_steps"]

        cur_users = self._get_current_concurrency(file)
        try:
            cur_users = int(cur_users)
        except ValueError:
            cur_users = 1

        users_list = deque()
        cur_time = round(self._now())

        if not rampup:
            # Ramp-up immediately
            rampup = 0
            steps = 1
        if not steps:
            steps = max(abs(users - cur_users), 1)
        ramp_in_sec = rampup * 60
        step_time = ramp_in_sec / steps
        users_pers_step = (users - cur_users)/steps

        prev_concurrency = cur_users
        for i in range(steps):
            concurrency = cur_users + (floor(users_pers_step*(i+1) + 0.5))
            if concurrency != prev_concurrency:
                users_list.append((concurrency, cur_time + step_time*i))
                prev_concurrency = concurrency
        self.log.info(f'Rampup plan: {users_list}')
        return users_list
